@@ -526,36 +526,6 @@ namespace pyopencl
 
 
   // context ------------------------------------------------------------------
-#define PYOPENCL_PARSE_CONTEXT_PROPERTIES \
-  cl_context_properties *props_ptr = 0; \
-  std::vector<cl_context_properties> props; \
-  \
-  if (py_properties.ptr() != Py_None) \
-  { \
-    PYTHON_FOREACH(prop_tuple, py_properties) \
-    { \
-      if (len(prop_tuple) != 2) \
-        throw error("Context", CL_INVALID_VALUE, "property tuple must have length 2"); \
-      cl_context_properties prop = \
-          py::extract<cl_context_properties>(prop_tuple[0]); \
-      props.push_back(prop); \
-   \
-      if (prop == CL_CONTEXT_PLATFORM) \
-      { \
-        py::extract<const platform &> value(prop_tuple[1]); \
-        props.push_back( \
-            reinterpret_cast<cl_context_properties>(value().data())); \
-      } \
-      else \
-        throw error("Context", CL_INVALID_VALUE, "invalid context property"); \
-    } \
-    props.push_back(0); \
-    props_ptr = &props.front(); \
-  }
-
-
-
-
   class context : public boost::noncopyable
   {
     private:
@@ -569,30 +539,6 @@ namespace pyopencl
           PYOPENCL_CALL_GUARDED(clRetainContext, (ctx));
       }
 
-      context(
-          py::list py_devices,
-          py::object py_properties=py::object())
-      {
-        PYOPENCL_PARSE_CONTEXT_PROPERTIES;
-
-        std::vector<cl_device_id> devices;
-        PYTHON_FOREACH(py_dev, py_devices)
-        {
-          py::extract<const device &> dev(py_dev);
-            devices.push_back(dev().data());
-        }
-
-        cl_int status_code;
-        m_context = clCreateContext(
-            props_ptr,
-            devices.size(),
-            &devices.front(),
-            0, 0, &status_code);
-
-        PYOPENCL_PRINT_CALL_TRACE("clCreateContext");
-        if (status_code != CL_SUCCESS)
-          throw pyopencl::error("Context", status_code);
-      }
 
       ~context()
       {
@@ -672,21 +618,76 @@ namespace pyopencl
 
 
 
-  context *create_context_from_type(
-      cl_device_type dev_type,
-      py::object py_properties)
+  context *create_context(py::object py_devices, py::object py_properties,
+      py::object py_dev_type)
   {
-    PYOPENCL_PARSE_CONTEXT_PROPERTIES;
+    // parse context properties
+    cl_context_properties *props_ptr = 0;
+    std::vector<cl_context_properties> props;
+  
+    if (py_properties.ptr() != Py_None)
+    {
+      PYTHON_FOREACH(prop_tuple, py_properties)
+      {
+        if (len(prop_tuple) != 2)
+          throw error("Context", CL_INVALID_VALUE, "property tuple must have length 2");
+        cl_context_properties prop =
+            py::extract<cl_context_properties>(prop_tuple[0]);
+        props.push_back(prop);
+    
+        if (prop == CL_CONTEXT_PLATFORM)
+        {
+          py::extract<const platform &> value(prop_tuple[1]);
+          props.push_back(
+              reinterpret_cast<cl_context_properties>(value().data()));
+        }
+        else
+          throw error("Context", CL_INVALID_VALUE, "invalid context property");
+      }
+      props.push_back(0);
+      props_ptr = &props.front();
+    }
 
     cl_int status_code;
-    cl_context ctx = clCreateContextFromType(
-        props_ptr,
-        dev_type,
-        0, 0, &status_code);
 
-    PYOPENCL_PRINT_CALL_TRACE("clCreateContextFromType");
+    cl_context ctx;
+
+    // from device list
+    if (py_devices.ptr() != Py_None)
+    {
+      if (py_dev_type.ptr() != Py_None)
+        throw error("Context", CL_INVALID_VALUE, 
+            "one of 'devices' or 'dev_type' must be None");
+
+      std::vector<cl_device_id> devices;
+      PYTHON_FOREACH(py_dev, py_devices)
+      {
+        py::extract<const device &> dev(py_dev);
+          devices.push_back(dev().data());
+      }
+
+      ctx = clCreateContext(
+          props_ptr,
+          devices.size(),
+          &devices.front(),
+          0, 0, &status_code);
+
+      PYOPENCL_PRINT_CALL_TRACE("clCreateContext");
+    }
+    // from dev_type
+    else
+    {
+      cl_device_type dev_type = CL_DEVICE_TYPE_DEFAULT;
+      if (py_dev_type.ptr() != Py_None)
+        dev_type = py::extract<cl_device_type>(py_dev_type)();
+
+      ctx = clCreateContextFromType(props_ptr, dev_type, 0, 0, &status_code);
+
+      PYOPENCL_PRINT_CALL_TRACE("clCreateContextFromType");
+    }
+
     if (status_code != CL_SUCCESS)
-      throw pyopencl::error("clCreateContextFromType", status_code);
+      throw pyopencl::error("Context", status_code);
 
     try
     {
@@ -694,13 +695,14 @@ namespace pyopencl
     }
     catch (...)
     {
-      PYOPENCL_CALL_GUARDED_CLEANUP(clReleaseContext, (ctx));
+      PYOPENCL_CALL_GUARDED(clReleaseContext, (ctx));
       throw;
     }
   }
 
 
 
+
   // command_queue ------------------------------------------------------------
   class command_queue
   {
@@ -963,17 +965,19 @@ namespace pyopencl
         if (!m_valid)
             throw error("MemoryObject.free", CL_INVALID_VALUE,
                 "trying to double-unref mem object");
-        PYOPENCL_CALL_GUARDED_CLEANUP(clReleaseMemObject,
-            (m_mem));
+        PYOPENCL_CALL_GUARDED_CLEANUP(clReleaseMemObject, (m_mem));
         m_valid = false;
       }
 
-      ~memory_object()
+      virtual ~memory_object()
       {
         if (m_valid)
           release();
       }
 
+      py::object hostbuf()
+      { return m_hostbuf; }
+
       const cl_mem data() const
       { return m_mem; }
 
@@ -1044,6 +1048,11 @@ namespace pyopencl
       py::object buffer
       )
   {
+    if (buffer.ptr() != Py_None && 
+        !(flags & (CL_MEM_USE_HOST_PTR | CL_MEM_COPY_HOST_PTR)))
+      PyErr_Warn(PyExc_UserWarning, "'hostbuf' was passed, "
+          "but no memory flags to make use of it.");
+
     void *buf = 0;
     py::object *retained_buf_obj = 0;
     if (buffer.ptr() != Py_None)
@@ -1098,9 +1107,16 @@ namespace pyopencl
       py::object buffer,
       size_t device_offset,
       py::object py_wait_for,
-      bool is_blocking
+      bool is_blocking,
+      py::object host_buffer_deprecated
       )
   {
+    if (host_buffer_deprecated.ptr() != Py_None)
+    {
+      PYOPENCL_DEPRECATED("The 'host_buffer' keyword argument", "0.93", );
+      buffer = host_buffer_deprecated;
+    }
+
     PYOPENCL_PARSE_WAIT_FOR;
 
     void *buf;
@@ -1129,9 +1145,16 @@ namespace pyopencl
       py::object buffer,
       size_t device_offset,
       py::object py_wait_for,
-      bool is_blocking
+      bool is_blocking,
+      py::object host_buffer_deprecated
       )
   {
+    if (host_buffer_deprecated.ptr() != Py_None)
+    {
+      PYOPENCL_DEPRECATED("The 'host_buffer' keyword argument", "0.93", );
+      buffer = host_buffer_deprecated;
+    }
+
     PYOPENCL_PARSE_WAIT_FOR;
 
     const void *buf;
@@ -1199,14 +1222,17 @@ namespace pyopencl
       cl_mem_flags flags,
       cl_mem_object_type image_type)
   {
-    cl_uint num_image_formats;
-    PYOPENCL_CALL_GUARDED(clGetSupportedImageFormats, (
-          ctx.data(), flags, image_type, 0, 0, &num_image_formats));
-
+    cl_uint num_image_formats = 1024;
     std::vector<cl_image_format> formats(num_image_formats);
-    PYOPENCL_CALL_GUARDED(clGetSupportedImageFormats, (
-          ctx.data(), flags, image_type,
-          num_image_formats, &formats.front(), 0));
+    do
+    {
+      if (formats.size() < num_image_formats)
+        formats.resize(num_image_formats);
+
+      PYOPENCL_CALL_GUARDED(clGetSupportedImageFormats, (
+            ctx.data(), flags, image_type,
+            formats.size(), &formats.front(), &num_image_formats));
+    } while (formats.size() < num_image_formats);
 
     PYOPENCL_RETURN_VECTOR(cl_image_format, formats);
   }
@@ -1220,8 +1246,30 @@ namespace pyopencl
       cl_image_format const &fmt,
       py::object shape,
       py::object pitches,
-      py::object buffer)
+      py::object buffer,
+      py::object host_buffer_deprecated
+      )
   {
+    if (host_buffer_deprecated.ptr() != Py_None)
+    {
+      PYOPENCL_DEPRECATED("The 'host_buffer' keyword argument", "0.93", );
+      buffer = host_buffer_deprecated;
+    }
+
+    if (buffer.ptr() != Py_None && 
+        !(flags & (CL_MEM_USE_HOST_PTR | CL_MEM_COPY_HOST_PTR)))
+      PyErr_Warn(PyExc_UserWarning, "'hostbuf' was passed, "
+          "but no memory flags to make use of it.");
+
+    if (shape.ptr() == Py_None)
+    {
+      if (buffer.ptr() == Py_None)
+        throw pyopencl::error("Image", CL_INVALID_VALUE, 
+            "'shape' must be passed if 'hostbuf' is not given");
+
+      shape = buffer.attr("shape");
+    }
+
     void *buf = 0;
     PYOPENCL_BUFFER_SIZE_T len;
     py::object *retained_buf_obj = 0;
@@ -1319,9 +1367,16 @@ namespace pyopencl
       py::object buffer,
       size_t row_pitch, size_t slice_pitch,
       py::object py_wait_for,
-      bool is_blocking
+      bool is_blocking,
+      py::object host_buffer_deprecated
       )
   {
+    if (host_buffer_deprecated.ptr() != Py_None)
+    {
+      PYOPENCL_DEPRECATED("The 'host_buffer' keyword argument", "0.93", );
+      buffer = host_buffer_deprecated;
+    }
+
     PYOPENCL_PARSE_WAIT_FOR;
     COPY_PY_COORD_TRIPLE(origin);
     COPY_PY_REGION_TRIPLE(region);
@@ -1353,9 +1408,16 @@ namespace pyopencl
       py::object buffer,
       size_t row_pitch, size_t slice_pitch,
       py::object py_wait_for,
-      bool is_blocking
+      bool is_blocking,
+      py::object host_buffer_deprecated
       )
   {
+    if (host_buffer_deprecated.ptr() != Py_None)
+    {
+      PYOPENCL_DEPRECATED("The 'host_buffer' keyword argument", "0.93", );
+      buffer = host_buffer_deprecated;
+    }
+
     PYOPENCL_PARSE_WAIT_FOR;
     COPY_PY_COORD_TRIPLE(origin);
     COPY_PY_REGION_TRIPLE(region);
@@ -1812,7 +1874,7 @@ namespace pyopencl
         }
       }
 
-      program &build(std::string options, py::object py_devices)
+      void build(std::string options, py::object py_devices)
       {
         if (py_devices.ptr() == Py_None)
         {
@@ -1829,8 +1891,6 @@ namespace pyopencl
               (m_program, devices.size(), &devices.front(),
                options.c_str(), 0 ,0));
         }
-
-        return *this;
       }
   };
 
@@ -1926,6 +1986,23 @@ namespace pyopencl
 
 
   // kernel -------------------------------------------------------------------
+  class local_memory
+  {
+    private:
+      size_t m_size;
+
+    public:
+      local_memory(size_t size)
+        : m_size(size)
+      { }
+
+      size_t size() const
+      { return m_size; }
+  };
+
+
+
+
   class kernel : boost::noncopyable
   {
     private:
@@ -1967,6 +2044,11 @@ namespace pyopencl
 
       PYOPENCL_EQUALITY_TESTS(kernel);
 
+      void set_arg_null(cl_uint arg_index)
+      {
+        PYOPENCL_CALL_GUARDED(clSetKernelArg, (m_kernel, arg_index, 0, 0));
+      }
+
       void set_arg_mem(cl_uint arg_index, memory_object &mo)
       {
         cl_mem m = mo.data();
@@ -1974,7 +2056,13 @@ namespace pyopencl
             (m_kernel, arg_index, sizeof(cl_mem), &m));
       }
 
-      void set_arg_sampler(cl_uint arg_index, sampler &smp)
+      void set_arg_local(cl_uint arg_index, local_memory const &loc)
+      {
+        PYOPENCL_CALL_GUARDED(clSetKernelArg,
+            (m_kernel, arg_index, loc.size(), 0));
+      }
+
+      void set_arg_sampler(cl_uint arg_index, sampler const &smp)
       {
         cl_sampler s = smp.data();
         PYOPENCL_CALL_GUARDED(clSetKernelArg,
@@ -1987,12 +2075,20 @@ namespace pyopencl
         PYOPENCL_BUFFER_SIZE_T len;
 
         if (PyObject_AsReadBuffer(py_buffer.ptr(), &buf, &len))
+          throw py::error_already_set();
+
         PYOPENCL_CALL_GUARDED(clSetKernelArg,
             (m_kernel, arg_index, len, buf));
       }
 
       void set_arg(cl_uint arg_index, py::object arg)
       {
+        if (arg.ptr() == Py_None)
+        {
+          set_arg_null(arg_index);
+          return;
+        }
+
         py::extract<memory_object &> ex_mo(arg);
         if (ex_mo.check())
         {
@@ -2000,7 +2096,14 @@ namespace pyopencl
           return;
         }
 
-        py::extract<sampler &> ex_smp(arg);
+        py::extract<local_memory const &> ex_loc(arg);
+        if (ex_loc.check())
+        {
+          set_arg_local(arg_index, ex_loc());
+          return;
+        }
+
+        py::extract<sampler const &> ex_smp(arg);
         if (ex_smp.check())
         {
           set_arg_sampler(arg_index, ex_smp());
@@ -2063,7 +2166,6 @@ namespace pyopencl
 
 
 
-
   py::list create_kernels_in_program(program &pgm)
   {
     py::list result;
@@ -2083,8 +2185,6 @@ namespace pyopencl
 
 
 
-
-
   event *enqueue_nd_range_kernel(
       command_queue &cq,
       kernel &knl,
